/* Symbol table support
   Copyright (C) 1998-99 James Bowman, Scott Dattalo

This file is part of gpasm.

gpasm is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpasm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "stdhdr.h"

#include "gpasm.h"
#include "symbol.h"
#include "lst.h"
#include "cod.h"
static int check =0;

static char * dir_block;        /* Directory cod block */
static char * cod_block;        /* General purpose cod block */
static char * cod_image_blocks[COD_CODE_IMAGE_BLOCKS];
static int last_cod_block = 1;
static int cod_lst_line_number = 0;
static enum cod_block_types cod_block_owner = cb_nobody;

/************************************************************************/

// a little utility routine that's not actually used in the
// program, but was useful in computing offsets into structures.

unsigned int ptr_offset(char *a, char *b)
{
  unsigned int c,d;
  c = (unsigned int) a;
  d = (unsigned int) b;

  if(d>c) d = (unsigned int) dir_block;

  if(c > (d+0x200))
    printf("******************** Buffer overflow\n");
  return( c-d);

}

/* copy a string to a cod block using the pascal convention, 
 * i.e. the string length occupies the first string location
 */

void cod_strncpy(char *dest, char *src, int max_len)
{


  *(dest-1) = ( (max_len>strlen(src)) ? strlen(src) : max_len );
  strncpy(dest, src, *(dest-1));

}

/*
 * get_short_int, put_short_int, and put_int
 * utility routines for accessing the cod blocks (char *)
 * using the proper endianess. (Using a union would be
 * cleaner, but less portable.
 */

int get_short_int(char * buff)
{
  return ( buff[0] + (buff[1] << 8));
}

void put_short_int(char * buff, int value)
{

  buff[0] = value & 0xff;
  buff[1] = (value >> 8) & 0xff;
}

void put_int(char * buff, int value)
{

  buff[0] = value & 0xff;
  buff[1] = (value >> 8) & 0xff;
  buff[2] = (value >> 16) & 0xff;
  buff[3] = (value >> 24) & 0xff;
}

// reverse the endian type (big to little or little to big)
// note: there's probably a wonderful macro nestled deep in
// the bowels of the os that nicely handles this conversion...

unsigned int endian_swap(unsigned int a)
{

  return
    ( (a>>24) & 0x000000ff) | 
    ( (a>>8)  & 0x0000ff00) |
    ( (a<<8)  & 0x00ff0000) |
    ( (a<<24) & 0xff000000);

}


/*
 * clear_cod_block - write zeroes to a code block, unless the
 * code block ptr is null.
 */
void clear_cod_block(char * block)
{
  int i;

  if(block)
    for(i=0; i<COD_BLOCK_SIZE; i++)
      block[i] = 0;
  else
    assert(0);
}

/*
 * init_cod_dir_block - initialize the directory block
 */

void init_cod_dir_block(void)
{
  int i;

  if(!state.cod.enabled)
    return;

  dir_block = malloc(COD_BLOCK_SIZE);
  cod_block = malloc(COD_BLOCK_SIZE);

  clear_cod_block(dir_block);
  clear_cod_block(cod_block);

  cod_block_owner = cb_list;

  for(i=0; i<COD_CODE_IMAGE_BLOCKS; i++)
    cod_image_blocks[i] = NULL;

  /* Initialize the directory block with known data. It'll be written
   * to the .cod file after everything else */

  cod_strncpy(&dir_block[COD_DIR_VERSION], 
	      GPASM_VERSION_STRING,
	      COD_DIR_COMPILER - COD_DIR_VERSION);
  cod_strncpy(&dir_block[COD_DIR_COMPILER], 
	      "gpasm",
	      COD_DIR_NOTICE - COD_DIR_COMPILER);
  cod_strncpy(&dir_block[COD_DIR_NOTICE], 
	      "Copyright (c) 1998-99 gnupic project",
	      COD_DIR_SYMTAB - COD_DIR_NOTICE);


  /* Move the cod file ptr to the begining of the second block */
  fseek(state.cod.f,COD_BLOCK_SIZE, SEEK_SET);
  last_cod_block = 1;

}

/*
 * write_cod_block - write a cod block to the .cod file and adjust
 * the cod block ptrs in the directory block.
 */

void write_cod_block(int block_ptr_start, int block_ptr_end, char *block)
{

  /* most of the cod blocks have a 'start' and 'end' pointer in the
   * directory block. These pointers are 16 bits wide. If the start
   * pointer is zero, then this is the first time a block of this
   * type has been written. In this case, the current block pointer
   * (stored in last_cod_block) is written to both the start and end 
   * pointer locations. 
   */

  if(!get_short_int(&dir_block[block_ptr_start]))
    put_short_int(&dir_block[block_ptr_start], last_cod_block);

  put_short_int(&dir_block[block_ptr_end], last_cod_block);

  fseek(state.cod.f,COD_BLOCK_SIZE*last_cod_block, SEEK_SET);
  fwrite(block, 1, COD_BLOCK_SIZE, state.cod.f);

  last_cod_block++;
}

/*
 * write_file_block - write a code block that contains a list of the
 * source files.
 */
void write_file_block(void)
{

#define FILES_PER_BLOCK COD_BLOCK_SIZE/COD_FILE_SIZE

  struct file_context *fc;

  clear_cod_block(cod_block);
  fc = state.files;       /* Get a pointer to the top of the file stack */

  while(fc != NULL) {

    /* The file id is used to define the index at which the file
     * name is written within the file code block. (The id's are
     * sequentially assigned when the files are opened.) If there
     * are too many files, then gpasm will abort. note: .cod files 
     * can handle larger file lists...
     */

    if(fc->id < FILES_PER_BLOCK)
      cod_strncpy(&cod_block[1 + COD_FILE_SIZE * fc->id],
		  fc->name,
		  COD_FILE_SIZE-1);
    else
      assert(0);

    fc = fc->prev;

  } 

  if(state.files)
    write_cod_block(COD_DIR_NAMTAB,COD_DIR_NAMTAB+2,cod_block);


}

/* cod_lst_line - add a line of information that cross references the
 * the opcode's address, the source file, and the list file.
 */

void cod_lst_line(void)
{
#define COD_LST_FIRST_LINE  7
  unsigned char smod_flag = 0xff;
  static int first_time = 1;
  int offset;

  if(!state.cod.enabled)
    return;

  /* Don't start until after the source is open or if lst_line doesn't
   * have posession of the cod_block  */
  if((state.src == NULL) | (cod_block_owner != cb_list))
    return;
    
  if(state.lst.line_number < COD_LST_FIRST_LINE)
    return;

  if(cod_lst_line_number >= COD_MAX_LINE_SYM) {
    write_cod_block(COD_DIR_LSTTAB,COD_DIR_LSTTAB+2,cod_block);
    cod_lst_line_number = 0;
    clear_cod_block(cod_block);
  }

  offset = cod_lst_line_number++ * COD_LINE_SYM_SIZE;
  cod_block[offset + COD_LS_SFILE] = state.src->fc->id;

  if(state.cod.emitting != 0)
    smod_flag = 0x080;
  else
    smod_flag = 0x90;

  if(first_time != 0) {
    first_time = 0;
    smod_flag  = 0xff;
  }

  cod_block[offset + COD_LS_SMOD]  = smod_flag;

  /* Write the source file line number corresponding to the list file line number */
  put_short_int(&cod_block[offset + COD_LS_SLINE], state.src->line_number);
  /*
  put_short_int(&cod_block[offset + COD_LS_SLINE], 
		((state.src->line_number) ? (state.src->line_number - 1) : 0));
  */

  /* Write the address of the opcode. Note: the '-1' takes into account that
   * the address is incremented previous to calling this routine */
  /*put_short_int(&cod_block[offset + COD_LS_SLOC], state.org);*/
  put_short_int(&cod_block[offset + COD_LS_SLOC],state.lst.line.was_org );

  /*  put_short_int(&cod_block[offset + COD_LS_SLOC],  
		((state.org) ? (state.org-1) : 0) );
  */
  /*  printf("***cod org %04x, srcline %05d flags %02x\n",state.lst.line.was_org,state.src->line_number,smod_flag); */

}

/* cod_write_symbols - write the symbol table to the .cod file
 */

void cod_write_symbols(struct symbol **symbol_list, int num_symbols)
{
  /* Each symbol is written as a dynamically sized structure to the 
   * .cod file. Its format is like this:
   * position  0               length of the symbol name
   * positions 1 to len        the symbol name
   * positions len+1 & len+2   Type of symbol (16 bits)
   * positions len+3 to len+7  Value of symbol
   */
#define COD_SYM_TYPE  1   // type info is 1 byte after the length
#define COD_SYM_VALUE 3   // value info is 3 bytes after the length
#define COD_SYM_EXTRA 7   // symbol name length + 7 is total structure size
#define MAX_SYM_LEN   255 // Maximum length of a symbol name

  int i,offset,len,type;
  struct variable *var;
  char * s;

  if(!state.cod.enabled)
    return;

  if(cod_block_owner == cb_list)
    write_cod_block(COD_DIR_LSTTAB,COD_DIR_LSTTAB+2,cod_block);

  cod_block_owner = cb_lsymbols;

  clear_cod_block(cod_block);
  offset = 0;

  for(i=0; i<num_symbols; i++) {
    var = get_symbol_annotation(symbol_list[i]);
    s = get_symbol_name(symbol_list[i]);
    len = strlen(s);

    /* If this symbol extends past the end of the cod block
     * then write this block out */

    if((offset + len + COD_SYM_EXTRA) >= COD_BLOCK_SIZE) {
      write_cod_block(COD_DIR_LSYMTAB,COD_DIR_LSYMTAB+2,cod_block);
      clear_cod_block(cod_block);
      offset = 0;
    }

    cod_strncpy(&cod_block[offset +1], s, MAX_SYM_LEN);

    if(var->type == gvt_cblock)
      type = COD_ST_C_SHORT;  // byte craft's nomenclature for a memory byte.
    else if(var->type == gvt_address)
      type = COD_ST_ADDRESS;
    else 
      type = COD_ST_CONSTANT;
    
    put_short_int(&cod_block[offset+len+COD_SYM_TYPE], type);
    put_int(&cod_block[offset+len+COD_SYM_VALUE], endian_swap(var->value));
    offset += (len+COD_SYM_EXTRA);
  }

  if(offset)
    write_cod_block(COD_DIR_LSYMTAB,COD_DIR_LSYMTAB+2,cod_block);

 
}

/* cod_emit_opcode - write one opcode to a cod_image_block
 */
void  cod_emit_opcode(int address,int opcode)
{

  int block_index;
  char * block;

  if(!state.cod.enabled)
    return;

  /* The code image blocks are handled in a different manner than the
   * other cod blocks. In theory, it's possible to emit opcodes in a
   * non-sequential manner. Furthermore, it's possible that there may
   * be gaps in the program memory. These cases are handled by an array
   * of code blocks. The lower 8 bits of the opcode's address form an
   * index into the code block, while bits 9-15 are an index into the
   * array of code blocks. The code image blocks are not written until
   * all of the opcodes have been emitted.
   */

  block_index = (address >> (COD_BLOCK_BITS-1)) & (COD_CODE_IMAGE_BLOCKS -1); 
  if(cod_image_blocks[block_index] == NULL) {
    cod_image_blocks[block_index] = malloc(COD_BLOCK_SIZE);
    clear_cod_block(cod_image_blocks[block_index]);
  }

  block = cod_image_blocks[block_index];

  put_short_int(&block[(address*2) & (COD_BLOCK_SIZE - 1)], opcode);

}

/* cod_write_range - record the starting and ending address of a
 * block of pic code. (E.g. if the pic source has code in different
 * sections of memory because of 'org' statements, then this is where
 * the .cod file records the start and end of those segments 
 */

void cod_write_range(int start, int end)
{

}
/* cod_write_code - write all of the assembled pic code to the .cod file
 */
void cod_write_code(void)
{
  int i,offset;
  int range, used_flag = 0;

  cod_block_owner = cb_range;

  clear_cod_block(cod_block);
  offset = 0;

  for(i=0; i<MAX_I_MEM; i++) {
    if (state.i_memory[i] & MEM_USED_MASK) {
      cod_emit_opcode(i, state.i_memory[i] & 0xffff);
      if(used_flag == 0) {
	range = i;       /* Save the start address in a range of opcodes */
	used_flag = 1;
      }
    } else {
      if(used_flag == 1) {
	put_short_int(&cod_block[offset], 2*range);
	put_short_int(&cod_block[offset+2], 2*(i-1));
	offset += 4;
	if(offset>=COD_BLOCK_SIZE) {
	  write_cod_block(COD_DIR_MEMMAP,COD_DIR_MEMMAP+2,cod_block);
	  offset = 0;
	}
	used_flag = 0;
      }
    }
    
  }

  if(offset)	
    write_cod_block(COD_DIR_MEMMAP,COD_DIR_MEMMAP+2,cod_block);

  /* write the code image blocks */
  for(i=0; i<COD_CODE_IMAGE_BLOCKS; i++)
    if(cod_image_blocks[i]) {
      write_cod_block(COD_DIR_CODE+i*2,COD_DIR_CODE+i*2,cod_image_blocks[i]);
      free(cod_image_blocks[i]);
    }
}

/*
 */
void cod_close_file(void)
{
  int i;

  if(!state.cod.enabled)
    return;

  if(cod_block_owner == cb_list)
    write_cod_block(COD_DIR_LSTTAB,COD_DIR_LSTTAB+2,cod_block);
  write_file_block();

  cod_write_code();

  cod_strncpy(&dir_block[COD_DIR_PROCESSOR], 
	      state.processor_info->names[2],
	      COD_DIR_LSYMTAB - COD_DIR_PROCESSOR);
  fseek(state.cod.f,0, SEEK_SET);
  fwrite(dir_block, 1, COD_BLOCK_SIZE, state.cod.f);

  fclose(state.cod.f);

  free(cod_block);
  free(dir_block);
  cod_block_owner == cb_nobody;
}

